#############
# $Id: harold_calibration.pm 12818 2011-03-16 12:53:22Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2010-03-05
# Last Changed On: $Date: 2011-03-16 12:53:22 +0000 (Wed, 16 Mar 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-26.1/lib/npg_pipeline/analysis/harold_calibration.pm $

package npg_pipeline::analysis::harold_calibration;
use Moose;
use MooseX::InsideOut;
use Carp;
use English qw{-no_match_vars};
use Cwd;      # for getcwd
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12818 $ =~ /(\d+)/mxs; $r; };
use npg_common::VersionComparison;

extends qw{npg_pipeline::analysis::bustard};

with qw{
  npg_pipeline::roles::business::harold_calibration_reqs
  npg_common::roles::run::intensities::config
  npg_pipeline::roles::business::pre_exec_strings
};

Readonly::Scalar our $PB_CAL_PIPELINE_SCRIPT => q{npg_pipeline_PB_cal};
Readonly::Scalar our $NEW_RTA_VERSION_GREATER_THAN_CURRENT_RUN_VERSION => -1;

=head1 NAME

npg_pipeline::analysis::harold_calibration

=head1 VERSION

$LastChangedRevision: 12818 $

=head1 SYNOPSIS

  my $oHaroldCalibration = npg_pipeline:analysis::harold_calibration->new();

=head1 DESCRIPTION

Object runner to launch internal calibration instead of CASAVA based calibration

=head1 SUBROUTINES/METHODS

=head2 launch_pb_cal

submits a job which will launch a separate pipeline process for the PB_cal calibration

  my $aJobIds = $oHaroldCalibration->launch_pb_cal({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

sub launch_pb_cal {
  my ( $self, $arg_refs ) = @_;

#  if ( $self->no_eland() || $self->no_recalibration() ) {
#    $self->log( q{This has been set to run with no recalibration step, I will not submit PB_cal pipeline launcher} );
#    return ();
#  }

  return ( $self->submit_bsub_command( $self->_pb_cal_command( $arg_refs ) ) );
}

# pb_cal_command generation
sub _pb_cal_command {
  my ( $self, $arg_refs ) = @_;

  my $job_dependencies = $arg_refs->{required_job_completion};

  my $intensity_path = $self->intensity_path();

  my $options = $self->attributes_as_command_options({
    excluded_attributes => [qw(
        log_file log_file_name log_file_path subpath casava_pipeline illumina_pipeline timestamp
        bustard_home gerald_config_prefix runtype script_name time recalibrated_path bustard_path pb_cal_path
        bcl_output_prefix bcl_executable make_output_prefix pb_cal_log_file basecall_path function_order
      )],
  });

  if ( $arg_refs->{only_basecall} ) {
    if ( $options !~ /--no_recalibration/xms ) {
      $options .= q{ --no_recalibration };
    }
  }

  # we need new pb_cal bustard and recalibrated equivalent paths
  my $time_now = $self->timestamp();
  my $basecall_path = $intensity_path . q{/PB_basecalls_} . $time_now;
  my $pb_cal_path   = $basecall_path;
  if ( $arg_refs->{only_basecall} || $self->no_recalibration() ) {
    $pb_cal_path .= q{/no_cal};
  } else {
    $pb_cal_path .= q{/PB_cal};
  }

  # make the new basecall and pb_cal directories with log directories in them
  $self->make_log_dir( $basecall_path );
  $self->make_log_dir( $pb_cal_path );

  my $job_command = $PB_CAL_PIPELINE_SCRIPT . qq{ $options};
  # doing these ensures the correct paths are posted through
  $job_command .= qq{ --bustard_path $basecall_path --basecall_path $basecall_path --pb_cal_path $pb_cal_path --recalibrated_path $pb_cal_path};

  my $job_name = $PB_CAL_PIPELINE_SCRIPT . q{_} . $self->id_run() . q{_} . $self->timestamp();

  my $bsub_command = q{bsub -q } . $self->lsf_queue() . q{ };# . $job_dependencies;
  $bsub_command .= q{ -J } . $job_name;
  $bsub_command .= q{ } . ( $self->fs_resource_string( {
    counter_slots_per_job => 1,
  } ) );
  $bsub_command .= q{ -o } . $self->runfolder_path() . q{/} . $job_name . q{.out};
  $bsub_command .= q{ '} . $job_command . q{'};

  if ( $self->verbose() ) {
    $self->log( $bsub_command );
  }

  return $bsub_command
}

=head2 run_bcl_setup

Run setupBclToQseq.py to create a makefile in the Basecalls directory
which creates qseq files from bcl files. This takes only a few seconds
to run and has no dependencies, so we just do it, rather than bsub it,
and we'll return an empty job_id array.

  my $aJobIds = $oHaroldCalibration->run_bcl_setup({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

has 'bcl_executable'    => ( is => 'ro', isa => 'Str', lazy_build => 1, );
has 'bcl_output_prefix' => ( is => 'ro', isa => 'Str', default => q{setupBclToQseq} );

sub _build_bcl_executable {
  my ( $self ) = @_;
  return $self->external_script_names_conf()->{bcl_to_qseq};
}

sub _bcl_pipeline {
  my ( $self ) = @_;

  my $bcl_olb = $self->illumina_pipeline_conf()->{bcl_olb};
  if ( $bcl_olb ) {
    $bcl_olb = $self->filesystem_locations_conf()->{src} . q{/} . $bcl_olb;
  }

  my $pipeline  = $bcl_olb || $self->illumina_pipeline();
  return $pipeline;
}

sub run_bcl_setup {

  my ($self, $arg_refs) = @_;

  my $job_dependencies = $arg_refs->{required_job_completion};
  my $job_ids = [];

  my $run_folder  = $self->runfolder_path();   # fully qualified name
  my $intensities = $self->intensity_path();
  my $basecalls   = $self->basecall_path();

  my $cur_dir = getcwd();                      # save for later
  if (!$ENV{TEST_BIN}) {
    chdir $basecalls or croak qq{could not cd to $basecalls};
  }

  my $bcl_exe   = $self->bcl_executable();
  my $timestamp = $self->timestamp();
  my $bcl_out   = qq{$basecalls/} . $self->bcl_output_prefix() . qq{-$timestamp.txt};

  my $final_flag = q{-c};
  if ( compare_version( $self->illumina_pipeline_conf()->{old_rta_version_below}, $self->rta_version() ) == $NEW_RTA_VERSION_GREATER_THAN_CURRENT_RUN_VERSION ) {
    $final_flag = qq{-f $intensities/BaseCalls};
  }

  my $pipeline = $self->_bcl_pipeline();
  my   @command;
  push @command, $self->_bcl_pipeline() . qq{/bin/$bcl_exe};     # TODO: Is this correct?
  push @command, qq{-p $intensities};
  push @command,  q{-P _pos.txt};
  push @command,  q{-i } . $intensities; # this is the location of the intensity files
  push @command,  q{-b } . $intensities . q{/BaseCalls}; # this is the location of the instrument created BaseCalls directory
  push @command, qq{-o $basecalls}; # output directory for qseq files
  push @command,  q{--overwrite --no-eamss}; # have removed the --in-place flag, since we use an output directory anyway
  push @command, $final_flag;
  push @command, qq{> $bcl_out};
  push @command,  q{2>&1};

  my $bcl_command = join q[ ], @command;
  $self->log( qq{BCL converter command: $bcl_command} );

  my $rc = system $bcl_command;
  croak qq{BCL converter command failed: $rc} if $rc;

  $self->log( q{BCL converter complete} );

  $self->log( qq{Changing back to $cur_dir} );
  chdir $cur_dir or croak qq{could not cd back to $cur_dir};  # go back home

  return @{ $job_ids };     # currently, an empty array

}

=head2 make_qseq_files

bsub a make job to create qseq files from the .bcl files in the
Basecalls directory, plus assorted plots and reports. The make file
was created in the run_bcl_setup step.

  my $aJobIds = $oHaroldCalibration->make_qseq_files({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

has 'make_output_prefix' => (is => 'ro', isa => 'Str', default => q{qseq_files});   # constant

sub make_qseq_files {

  my ($self, $arg_refs) = @_;

  my $job_dependencies = $arg_refs->{required_job_completion};
  my $job_ids = [];

  my $id_run      = $self->id_run();
  my $run_folder  = $self->runfolder_path();   # fully qualified name
  my $intensities = $self->intensity_path();
  my $basecalls   = $self->basecall_path();

  my $cur_dir = getcwd();                      # save for later
  if (!$ENV{TEST_BIN}) {
    chdir $basecalls or croak "could not cd to $basecalls";
  }

  my $timestamp   = $self->timestamp();
  my $prefix      = $self->make_output_prefix();
  my $job_name    = "${prefix}_${id_run}_all_${timestamp}";
  my $output_name = "log/${prefix}_${id_run}_all_${timestamp}-\%J.out";

  my $bsub_queue  = $self->lsf_queue;
  my $jobs        = $self->processor_fork_number();
  my $mem_req     = $self->memory_requirement();

  my   @command;
  push @command, 'bsub';
  push @command, "-n $jobs";
  push @command, "-q $bsub_queue";
  push @command, "-o $output_name ";
  push @command, "-J $job_name";
  push @command, '-C ' . $self->dump_size();    # get dump on failure

  push @command, $self->fs_resource_string( {
    resource_string => qq{-R 'select[mem>$mem_req] rusage[mem=$mem_req] span[hosts=1]'},
    counter_slots_per_job => $self->processor_fork_number(),
  });

  push @command, "-M ${mem_req}000";
  push @command, $job_dependencies || q[];
  push @command, $self->make_executable();
  push @command, "-j $jobs";

  my $bsub_command = join q[ ], @command;

  $self->log("make command: $bsub_command");

  my $job = $self->submit_bsub_command($bsub_command);

  $self->log( qq{Changing back to $cur_dir} );
  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return $job;

}

=head2 generate_alignment_files

submit the jobs which will generate bam alignment files ready to pass onto calibration table generator

  my $aJobIds = $oHaroldCalibration->generate_alignment_files({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

sub generate_alignment_files {
  my ( $self, $arg_refs ) = @_;
  my $job_ids = [];

  if ( $self->no_eland() || $self->no_recalibration() ) {
    $self->log( q{This has been set to run with no recalibration step} );
    return ();
  }

  my $job_dependencies = $arg_refs->{required_job_completion};

  my $run = $self->run();

  my $cur_dir = getcwd();            # save for later
  my $dir = $self->basecall_path();
  $self->make_log_dir( $dir ); # create a log directory within BaseCalls

  # create the calibration directory
  my $pb_cal_dir = $self->create_pb_calibration_directory();

  $self->_set_recalibrated_path( $self->pb_cal_path() );

  $self->log( qq{Changing to $pb_cal_dir} );
  chdir $pb_cal_dir or croak "could not cd to $pb_cal_dir"; # change to the calibration directory

  foreach my $run_lane ( @{$run->run_lanes()} ) {
    if ( $self->no_recalibration() ) {
      $self->log( q{no_recalibration option selected, no processing of calibration tables for any lanes performed} );
      last;
    }

    $self->_generate_alignment_file_per_lane( {
      run_lane => $run_lane,
      job_ids => $job_ids,
      dir => $dir,
      job_dependencies => $job_dependencies,
    } );
  }

  $self->log( qq{Changing back to $cur_dir} );
  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return @{ $job_ids };
}

=head2 generate_calibration_table

submit the bsub jobs which will create the calibration tables, returning an array of job_ids.

  my $aJobIds = $oHaroldCalibration->generate_calibration_table({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

sub generate_calibration_table {
  my ($self, $arg_refs) = @_;

  if ( $self->no_eland() || $self->no_recalibration() ) {
    $self->log( q{This has been set to run with no recalibration step} );
    return ();
  }

  my $job_dependencies = $arg_refs->{required_job_completion};

  my $run     = $self->run();
  my $job_ids = [];

  my $id_run = $self->id_run();

  my $cur_dir = getcwd();            # save for later
  my $dir = $self->basecall_path();

  # create the calibration directory
  my $pb_cal_dir = $self->create_pb_calibration_directory();

  $self->_set_recalibrated_path( $self->pb_cal_path() );

  $self->log( qq{Changing to $pb_cal_dir} );
  chdir $pb_cal_dir or croak "could not cd to $pb_cal_dir"; # change to the calibration directory

  foreach my $run_lane ( @{$run->run_lanes()} ) {
    if ( $self->no_recalibration() ) {
      $self->log( q{no_recalibration option selected, no processing of calibration tables for any lanes performed} );
      last;
    }

    $self->_generate_calibration_table_per_lane( {
      run_lane => $run_lane,
      job_ids => $job_ids,
      dir => $dir,
      job_dependencies => $job_dependencies,
    } );
  }

  $self->log( qq{Changing back to $cur_dir} );
  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return @{ $job_ids };
}

=head2 generate_recalibrated_qseqs

submit the bsub jobs which will recalibrate the lanes, returning an array of job_ids.

  my $aJobIds = $oHaroldCalibration->generate_recalibrated_qseqs({
    required_job_completion => $sJobRequirenmentString,
  });

=cut

sub generate_recalibrated_qseqs {
  my ($self, $arg_refs) = @_;

  if ( $self->no_eland() || $self->no_recalibration() ) {
    $self->log( q{This has been set to run with no recalibration step} );
    return ();
  }

  my $dir = $self->basecall_path();
  my $pb_cal_dir = $self->pb_cal_path();

  if ( ! $self->directory_exists($pb_cal_dir) ) {
    $self->log( qq{$pb_cal_dir does not exist, not executing jobs} );
    return ();
  }

  my $id_run = $self->id_run();

  my ( $read_1_calibration_table, $read_2_calibration_table );

  # if there is a control lane, then this will be used for any lanes with no reference sequence
  $read_1_calibration_table = $self->calibration_table_name( {
      id_run => $id_run,
      read => 1,
  } );

  $read_2_calibration_table = $self->calibration_table_name( {
    id_run => $id_run,
    read => 2,
  } );


  my $job_dependencies = $arg_refs->{required_job_completion};

  my $run     = $self->run();
  my $job_ids = [];

  my $cur_dir = getcwd();            # save for later

  $self->log( qq{Changing to $pb_cal_dir} );
  chdir $pb_cal_dir or croak "could not cd to $pb_cal_dir"; # change to the calibration directory

  foreach my $run_lane ( @{$run->run_lanes()} ) {
    my $arg_ref_hash = {
      read_1_calibration_table => $read_1_calibration_table,
      read_2_calibration_table => $read_2_calibration_table,
      job_ids => $job_ids,
      run_lane => $run_lane,
      dir => $dir,
      job_dependencies => $job_dependencies,
    };
    $self->_generate_recalibrated_qseqs_per_lane( $arg_ref_hash );
  }

  $self->log( qq{Changing back to $cur_dir} );
  chdir $cur_dir or croak "could not cd back to $cur_dir";  # go back home

  return @{ $job_ids };
}

##########
# private methods

sub _generate_recalibrated_qseqs_per_lane {
  my ( $self, $arg_refs ) = @_;

  my $id_run = $self->id_run();

  my $run_lane = $arg_refs->{run_lane};
  my $lane = $run_lane->position();

  my $job_ids = $arg_refs->{job_ids};

  my $cal_table_1_to_use = $arg_refs->{read_1_calibration_table};
  my $cal_table_2_to_use = $arg_refs->{read_2_calibration_table};

  my $ref_seq = $self->ref_seq( {
      lane => {
        id_run => $id_run,
        position => $lane,
      }
    } );

  # if there is a reference sequence, or the lane has spiked in phix, then use lanes own caltable
  if ( $ref_seq || $run_lane->is_spiked_phix() ) {
    $cal_table_1_to_use = $self->calibration_table_name( {
      id_run => $id_run,
      read => $self->general_values_conf()->{sanger_read_one},
      position => $lane,
    } );
  }

  my $mode;
  if ( $self->has_pb_cal_mode() ) {
    $mode = $self->pb_cal_mode();
  } elsif ( $cal_table_1_to_use eq $arg_refs->{read_1_calibration_table} ) {
    # using control lane recal table
    if ( $cal_table_1_to_use !~ /cycle/xms ) {
      $mode = 2;
    }
  } else {};

  my $clmode;
  if ( $arg_refs->{read_1_calibration_table} !~ /cycle/xms ) {
    $clmode = 2;
  }

  my $args_read1 = {
    dir => $arg_refs->{dir},
    position => $lane,
    read => $self->general_values_conf()->{sanger_read_one},
    job_dependencies => $arg_refs->{job_dependencies},
    calibration_table => $cal_table_1_to_use,
    clct => $arg_refs->{read_1_calibration_table},
    mode => $mode,
    clmode => $clmode,
  };

  my $args_read2 = {};

  if ( $self->is_paired_read() ) {

    # if there is a reference sequence, then use lanes own caltable
    if ( $ref_seq ) {
      my $args = {
        id_run => $id_run,
        read => $self->general_values_conf()->{sanger_read_two},
        position => $lane,
      };
      if ( $self->is_indexed() ) {
        $args->{read} = $self->general_values_conf()->{illumina_read_two};
      }
      $cal_table_2_to_use = $self->calibration_table_name( $args );
    }

    $args_read1->{clpct} = $arg_refs->{read_2_calibration_table};
    $args_read1->{pct} = $cal_table_2_to_use;

    if ( $arg_refs->{read_2_calibration_table} !~ /cycle/xms ) {
      $clmode = 2;
    }

    $args_read2 = {
      dir => $arg_refs->{dir},
      position => $lane,
      read => $self->general_values_conf()->{sanger_read_two},
      job_dependencies => $arg_refs->{job_dependencies},
      calibration_table => $cal_table_2_to_use,
      clct => $arg_refs->{read_2_calibration_table},
      clpct => $arg_refs->{read_1_calibration_table},
      pct => $cal_table_1_to_use,
      mode => $mode,
      clmode => $clmode,
    };

    if ( $self->is_indexed() ) {
      $args_read2->{read} = $self->general_values_conf()->{illumina_read_two};
    }

  }


  # only submit if there is a recalibration to do, either against a control lane or a reference
  if ( $ref_seq || $self->pick_control_lane() ) {

    my $bsub_command = $self->_recalibration_bsub_command( $args_read1 );
    if ( $self->verbose() ) {
      $self->log( $bsub_command );
    }

    push @{ $job_ids }, $self->submit_bsub_command( $bsub_command );

    if ( $self->is_paired_read() ) {
      $bsub_command = $self->_recalibration_bsub_command( $args_read2 );
      if ( $self->verbose() ) {
        $self->log( $bsub_command );
      }
      push @{ $job_ids }, $self->submit_bsub_command( $bsub_command );
    }

  }

  return;
}

sub _generate_calibration_table_per_lane {
  my ( $self, $arg_refs ) = @_;

  my $run_lane = $arg_refs->{run_lane};
  my $job_ids  = $arg_refs->{job_ids};
  my $id_run   = $self->id_run();

  my $library  = $run_lane->library();  # only one of $library, $control and $pool will be defined
  my $control  = $run_lane->control();
  my $pool     = $run_lane->pool();

  my $entity   = $run_lane->entity();   # entity is library || control || pool
  my $position = $run_lane->position();
  my $force_control = $self->has_control_lane_position() ? $self->control_lane_position() : 0;

  my $ref_seq;

  my $is_control = 0;
  my $is_spiked_phix = 0;
  # if the lane is a control, or has been declared the forced control
  if ( ! $self->no_control_lane() && ( $control || ( $force_control == $position ) )  ) {

    $ref_seq  = $self->ref_seq( {
      control => 1,
    } );
    $self->log( q{control - recalibrate control lane } . $position );
    $is_control = 1;

  # else see if the lane has a reference sequence
  } else {

    if ( $run_lane->is_spiked_phix() ) {
      $self->log( q{noncpntrol lane } . $position . q{ contains spiked-in phix} );
      $ref_seq = $self->ref_seq( {
        is_spiked_phix => 1,
      } );
      $is_spiked_phix++;
    } else {

      my $log_msg = q{control - noncontrol lane} . $position . qq{\n\tchecking to see if reference};
      $self->log( $log_msg );

      $ref_seq = $self->ref_seq( {
        lane => {
          id_run => $id_run,
          position => $position,
        }
      } ) || q{};

    }

    if ( $ref_seq ) {
      $self->log( qq{$id_run : $position : $ref_seq} );
    } else {
      $self->log( q{no ref seq found, no recalibration table generated for lane } . $position );
      return;
    }

  }

  if ( ! $ref_seq ) {
    return;
  }

  my $bsub_command = $self->_calibration_table_bsub_command( {
    dir => $arg_refs->{dir},
    position => $position,
    read => $self->general_values_conf()->{sanger_read_one},
    job_dependencies => $arg_refs->{job_dependencies},
    control => $is_control,
    is_spiked_phix => $is_spiked_phix,
  } );

  if ( $self->verbose() ) {
    $self->log( $bsub_command );
  }

  push @{ $job_ids }, $self->submit_bsub_command($bsub_command);

  if ( $self->is_paired_read() ) {
    my $args = {
      dir => $arg_refs->{dir},
      position => $position,
      read => $self->general_values_conf()->{sanger_read_two},
      job_dependencies => $arg_refs->{job_dependencies},
      control => $is_control,
      is_spiked_phix => $is_spiked_phix,
    };

    if ( $self->is_indexed() ) {
      $args->{read} = $self->general_values_conf()->{illumina_read_two};
    }

    $bsub_command = $self->_calibration_table_bsub_command( $args );

    if ( $self->verbose() ) {
      $self->log( $bsub_command );
    }

    push @{ $job_ids }, $self->submit_bsub_command($bsub_command);
  }

  return;
}

# generate the alignment file
sub _generate_alignment_file_per_lane {
  my ( $self, $arg_refs ) = @_;

  my $run_lane = $arg_refs->{run_lane};
  my $job_ids  = $arg_refs->{job_ids};
  my $id_run   = $self->id_run();

  my $library  = $run_lane->library();  # only one of $library, $control and $pool will be defined
  my $control  = $run_lane->control();
  my $pool     = $run_lane->pool();

  my $entity   = $run_lane->entity();   # entity is library || control || pool
  my $position = $run_lane->position();
  my $force_control = $self->has_control_lane_position() ? $self->control_lane_position() : 0;

  my $ref_seq;

  my $is_control = 0;
  my $is_spiked_phix = 0;

  # if the lane is a control, or has been declared the forced control
  if ( ! $self->no_control_lane() && ( $control || ( $force_control == $position ) )  ) {

    $ref_seq  = $self->ref_seq( {
      control => 1,
    } );
    $self->log( q{control - alignment file done with control reference } . $position );
    $is_control = 1;

  # else see if the lane has a reference sequence
  } else {

    if ( $run_lane->is_spiked_phix() ) {
      $self->log( q{noncpntrol lane } . $position . q{ contains spiked-in phix} );
      $ref_seq = $self->ref_seq( {
        is_spiked_phix => 1,
      } );
      $is_spiked_phix++;
    } else {

      my $log_msg = q{noncontrol lane} . $position . qq{\n\tchecking to see if reference};
      $self->log( $log_msg );

      $ref_seq = $self->ref_seq( {
        lane => {
          id_run => $id_run,
          position => $position,
        }
      } ) || q{};

    }

    if ( $ref_seq ) {
      $self->log( qq{$id_run : $position : $ref_seq} );
    } else {
      $self->log( q{no ref seq found, no bam alignment file generated for lane } . $position );
      return;
    }

  }

  if ( ! $ref_seq ) {
    return;
  }

  my $bsub_command = $self->_alignment_file_bsub_command( {
    dir => $arg_refs->{dir},
    position => $position,
    job_dependencies => $arg_refs->{job_dependencies},
    ref_seq => $ref_seq,
    control => $is_control,
    is_paired => $self->is_paired_read(),
    is_spiked_phix => $is_spiked_phix,
  } );

  if ( $self->verbose() ) {
    $self->log( $bsub_command );
  }

  push @{ $job_ids }, $self->submit_bsub_command($bsub_command);

  return;
}

# generate bsub command for generating the alignment files required
sub _alignment_file_bsub_command {
  my ( $self, $arg_refs ) = @_;

  my $position          = $arg_refs->{position};
  my $job_dependencies  = $arg_refs->{job_dependencies};
  my $ref_seq           = $arg_refs->{ref_seq};
  my $is_paired         = $arg_refs->{is_paired};
  my $is_spiked_phix    = $arg_refs->{is_spiked_phix};

  my $mem_size    = $self->general_values_conf()->{bam_creation_memory};
  my $timestamp   = $self->timestamp();
  my $bsub_queue  = $self->lsf_queue;
  my $id_run      = $self->id_run();

  my $job_name  = $self->is_paired_read() ? $self->align_job() . q{_} . $id_run . q{_} . $position . q{_paired_} . $timestamp
                :                           $self->align_job() . q{_} . $id_run . q{_} . $position . q{_} . $timestamp
                ;

  my @command;
  push @command, 'bsub';
  push @command, "-q $bsub_queue";
  push @command, $self->ref_adapter_pre_exec_string();
  push @command, q{-o log/}. $job_name . q{.out};

  push @command, "-J $job_name";
  push @command, '-C ' . $self->dump_size();    # get dump on failure

  push @command, $self->fs_resource_string( {
    resource_string => qq{-R 'select[mem>$mem_size] rusage[mem=$mem_size] span[hosts=1]'},
    counter_slots_per_job => 1,
  } );
  push @command, qq{-M ${mem_size}000};
  push @command, $job_dependencies || q[];

  push @command,  q["];               # " enclose command in quotes
  push @command, $self->pb_calibration_bin() . q{/} . $self->alignment_script();
  push @command, qq{--run $id_run};
  push @command, qq{--lane $position};
  if ( ! $is_spiked_phix ) {
    push @command,  q{--random } . $self->random();
  }
  push @command,  q{--runfolder_dir } . $self->runfolder_path();
  push @command,  q{--bustard_dir }   . $self->basecall_path();
  push @command,  q{--ref } . $ref_seq;
  push @command,  q{--} . $self->pb_cal_pipeline_conf()->{default_directory_for_qseqs};

  if ( $is_paired && $self->is_indexed() ) {
    push @command, q{--read1 } . $self->general_values_conf()->{sanger_read_one};
    push @command, q{--read2 } . $self->general_values_conf()->{illumina_read_two};
  } elsif ( $is_paired ) {
    push @command, q{--read1 } . $self->general_values_conf()->{sanger_read_one};
    push @command, q{--read2 } . $self->general_values_conf()->{sanger_read_two};
  } else {
    push @command, q{--read } . $self->general_values_conf()->{sanger_read_one};
  }

  push @command,  q{--pf_filter};

  push @command, q["];               # " closing quote

  my $bsub_command = join q[ ], @command;
  return $bsub_command;
}



# generate bsub command for recalibrating the lane qseq data
sub _recalibration_bsub_command {
  my ($self, $arg_refs) = @_;
  my $dir = $arg_refs->{dir};
  my $position = $arg_refs->{position};
  my $read = $arg_refs->{read};
  my $job_dependencies = $arg_refs->{job_dependencies};
  my $cal_table = $arg_refs->{calibration_table};

  my $mem_size = $self->mem_score();
  my $timestamp   = $self->timestamp();
  my $bsub_queue  = $self->lsf_queue;
  my $id_run = $self->id_run();

  my $job_name  = $self->is_paired_read() ? $self->score_job() . q{_} . $id_run . q{_} . $position . q{_} . $read . q{_} . $timestamp
                :                           $self->score_job() . q{_} . $id_run . q{_} . $position . q{_} . $timestamp
                ;

  my $cycle_start;
  if ($read == 1) {
    $cycle_start = 1;
  } else {
    my @r2r = $self->read2_cycle_range();
    $cycle_start = $r2r[0];
  }

  my @command;
  push @command, 'bsub';
  push @command, "-q $bsub_queue";
  push @command, q{-o log/}. $job_name . q{.out};

  push @command, "-J $job_name";
  push @command, '-C ' . $self->dump_size();    # get dump on failure

  push @command, $self->fs_resource_string( {
    resource_string => qq{-R 'select[mem>$mem_size] rusage[mem=$mem_size] span[hosts=1]'},
    counter_slots_per_job => 1,
  } );
  push @command, qq{-M ${mem_size}000};
  push @command, $job_dependencies || q[];

  push @command, q["];               # " enclose command in quotes
  push @command, $self->pb_calibration_bin() . q{/} . $self->recalibration_script();
  push @command,  q{--runfolder_dir } . $self->runfolder_path();
  push @command, qq{--bustard_dir $dir};
  push @command, qq{--run $id_run};
  push @command, qq{--lane $position};
  push @command, qq{--read $read};
  push @command, qq{--ct $cal_table};
  push @command, qq{--cstart $cycle_start};
  if ( $arg_refs->{mode} ) {
    push @command, q{--mode } . $arg_refs->{mode};
  }

  if ( $arg_refs->{pct} ) {
    push @command, q{--pct } . $arg_refs->{pct};
  }

  if ( $arg_refs->{clct} && $arg_refs->{clct} ne $cal_table ) {

    push @command, q{--clct } . $arg_refs->{clct};

    if ( $arg_refs->{clpct} ) {
      push @command, q{--clpct } . $arg_refs->{clpct};
    }

    if ( $arg_refs->{clmode} ) {
      push @command, q{--clmode } . $arg_refs->{clmode};
    } else {
      push @command, q{--clmode 2}; # assume 2, since most recalibration done with a control will be mode 2 (no per cycle)
    }
  }

  push @command, q["];               # " closing quote

  my $bsub_command = join q[ ], @command;
  return $bsub_command;
}

# generate bsub command for generating the calibration table required
sub _calibration_table_bsub_command {
  my ($self, $arg_refs) = @_;
  my $dir = $arg_refs->{dir};
  my $position = $arg_refs->{position};
  my $read = $arg_refs->{read};
  my $job_dependencies = $arg_refs->{job_dependencies};

  my $control = $arg_refs->{control};

  my $mem_size = $self->mem_calibration();
  my $timestamp   = $self->timestamp();
  my $bsub_queue  = $self->lsf_queue;
  my $id_run = $self->id_run();

  my $job_name  = $self->is_paired_read() ? $self->cal_table_job() . q{_} . $id_run . q{_} . $position . q{_} . $read . q{_} . $timestamp
                :                           $self->cal_table_job() . q{_} . $id_run . q{_} . $position . q{_} . $timestamp
                ;

  my $cycle_start;
  if ($read == 1) {
    $cycle_start = 1;
  } else {
    my @r2r = $self->read2_cycle_range();
    $cycle_start = $r2r[0];
  }

  my @command;
  push @command, 'bsub';
  push @command, "-q $bsub_queue";
  push @command, $self->ref_adapter_pre_exec_string();
  push @command, q{-o log/}. $job_name . q{.out};

  push @command, "-J $job_name";
  push @command, '-C ' . $self->dump_size();    # get dump on failure

  push @command, $self->fs_resource_string( {
    resource_string => qq{-R 'select[mem>$mem_size] rusage[mem=$mem_size] span[hosts=1]'},
    counter_slots_per_job => 1,
  } );
  push @command, qq{-M ${mem_size}000};
  push @command, $job_dependencies || q[];

  push @command,  q["];               # " enclose command in quotes
  push @command, $self->pb_calibration_bin() . q{/} . $self->cal_table_script();
  push @command,  q{--runfolder_dir } . $self->runfolder_path();
  push @command, qq{--bustard_dir $dir};
  push @command, qq{--run $id_run};
  push @command, qq{--lane $position};
  push @command,  q{--t_filter } . $self->t_filter();
  push @command, qq{--read $read};
  push @command, qq{--cstart $cycle_start};
  push @command,  q{--prefix } . $id_run . q{_} . $position . q{_} . $read;

  if ( $self->has_pb_cal_mode() ) {
    push @command, q{--mode } . $self->pb_cal_mode();
  } elsif ( $control ) {
    push @command, q{--mode 2};
  } else {};

  if ( $arg_refs->{is_spiked_phix} ) {
    push @command, q{--snp } . $self->pb_cal_pipeline_conf()->{phix_snp_file};
  }

  push @command, qq{--bam ${id_run}_${position}.bam};

  push @command, q["];               # " closing quote

  my $bsub_command = join q[ ], @command;
  return $bsub_command;
}

=head2 submit_post_qseq

Submit a job to invoke post-analysis processing (archival file
creation, QC, etc). All that job will do is another bsub. But this
extra level of indirection is necessary so that job dependencies can
be honoured. Returns array containing the job id

  my @JobIds = $oAnalysisBustard->submit_post_qseq({
    required_job_completion => $sRequiredJobCompletion, # an LSF -w'done(xxx) && ..' string
  });

=cut

sub submit_post_qseq {

  my ( $self, $arg_refs ) = @_;

  my $id_run = $self->id_run();
  $self->log("Submitting post-qseq job for run $id_run");

  my $bsub_command = $self->_post_qseq_cmd( $arg_refs );

  my $job = $self->submit_bsub_command( $bsub_command );

  return ( $job );

}

# ----------------------------------------------------------------------
# Separated out of the submission method for testing purposes.

sub _post_qseq_cmd {

  my ( $self, $arg_refs ) = @_;

  my $id_run      = $self->id_run();
  my $required_job_completion = $arg_refs->{required_job_completion};
  my $bsub_queue  = $self->lsf_queue;
  my $timestamp   = $self->timestamp();
  $self->_set_recalibrated_path( $self->pb_cal_path() );
  my $job_name    = q{post_qseq_} . $id_run . q{_} . $timestamp;
  my $output_name = q{'} . $self->pb_cal_path() . q{/log/}. $job_name . q{-%} . q{J.out'};

  # build the pb_cal_log_file so that this can be passed through to enable
  # the correct finding of the pipeline that was run
  $self->pb_cal_log_file();

  my $options = $self->attributes_as_command_options({
    excluded_attributes => [qw(
      log_file log_file_name log_file_path subpath casava_pipeline illumina_pipeline timestamp make_output_prefix
      bustard_home gerald_config_prefix runtype script_name time bcl_output_prefix bcl_executable
      bcl_output_prefix bcl_executable make_output_prefix function_order
      )],
  });

  my @command;
  push @command, q{bsub};
  push @command, qq{-q $bsub_queue};
  push @command, $required_job_completion || q[];
  push @command, $self->fs_resource_string( {
    counter_slots_per_job => 1,
  } );
  push @command, qq{-o $output_name};
  push @command, qq{-J $job_name};
  push @command, q{'} . $self->post_qseq_executable();

  push @command, $options . q{'};

  my $bsub_command = join q{ }, @command;

  $self->log( q{post-qseq command: } . $bsub_command );

  return $bsub_command;

}


no Moose;
__PACKAGE__->meta->make_immutable;
1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
