#############
# $Id: internal_info.pm 12818 2011-03-16 12:53:22Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-11-02
# Last Changed On: $Date: 2011-03-16 12:53:22 +0000 (Wed, 16 Mar 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-26.1/lib/npg_pipeline/roles/business/internal_info.pm $

package npg_pipeline::roles::business::internal_info;
use Moose::Role;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12818 $ =~ /(\d+)/mxs; $r; };

Readonly::Scalar our $MD5_CREATOR_SCRIPT          => q{md5sum};

Readonly::Scalar our $BSUB_OUTPUT_FILE  => q{substitute_this.out};
Readonly::Scalar our $BSUB_ERROR_FILE   => q{substitute_this.err};
Readonly::Scalar our $SUMMARY_LINK      => q{Latest_Summary};

=head1 NAME

npg_pipeline::roles::business::internal_info

=head1 VERSION

$LastChangedRevision: 12818 $

=head1 SYNOPSIS

  package MyPackage;
  use Moose;
  ...
  with qw{npg_pipeline::roles::business::internal_info};

=head1 DESCRIPTION

=head1 SUBROUTINES/METHODS

=head2 sanger_read_one

defaults to value for sanger_read_one from general_values.ini

=head2 sanger_read_two

defaults to value for sanger_read_two from general_values.ini

=head2 illumina_read_two

defaults to value for illumina_read_two from general_values.ini - applicable to use only if the run is indexed

=head2 sanger_read_index

defaults to value for sanger_read_index from general_values.ini

=head2 illumina_read_index

defaults to value for illumina_read_index from general_values.ini

These are all part of the qseq and fastq reading name applied to files being generated. The Sanger Institute
naming policy has the paired end named 2, and the index t, but the Illlumina software numbers according to the
point the run was done, which leaves indexed (multiplex) runs, with the tag read as 2 and the paired end 3

=cut

sub sanger_read_one     { my ( $self ) = @_; return $self->general_values_conf()->{sanger_read_one}; }
sub sanger_read_two     { my ( $self ) = @_; return $self->general_values_conf()->{sanger_read_two}; }
sub illumina_read_two   { my ( $self ) = @_; return $self->general_values_conf()->{illumina_read_two}; }
sub sanger_read_index   { my ( $self ) = @_; return $self->general_values_conf()->{sanger_read_index}; }
sub illumina_read_index { my ( $self ) = @_; return $self->general_values_conf()->{illumina_read_index}; }

=head2 default_usr_bin - the default usr bin on your system, defaults to value in filesystem_locations.ini

  my $sDefaultUsrBin = $class->default_usr_bin();

=cut

has q{default_usr_bin} => (isa => q{Str}, is => q{ro}, lazy_build => 1,
                           documentation => q{the default usr bin on your system - set if not using default from config},);


sub _build_default_usr_bin {
  my ($self) = @_;
  my $test_dir = q{};
  if ($ENV{TEST_BIN}) {
    $test_dir = $ENV{TEST_BIN};
  }
  return $test_dir . $self->filesystem_locations_conf()->{user_bin};
}


=head2 summary_link - the name used for the link to the recalibrated directory which is of the latest analysis from the run_folder
defaults to Latest_Summary

  my $sSummaryLink = $class->summary_link();

=cut

has q{summary_link} => (isa => q{Str}, is => q{ro}, lazy_build => 1,
                        documentation => qq{the name used for the link to the recalibrated directory which is of the latest analysis from the run_folder, defaults to $SUMMARY_LINK},);

sub _build_summary_link {
  my ($self) = @_;
  return $SUMMARY_LINK;
}

=head2 max_lane - the max lane number on a flowcell, defaults to value for max_illumina_lanes in general_values.ini

  my $iMaxLane = $class->max_lane();

=cut

has q{max_lane} => (isa => q{Int}, is => q{ro}, lazy_build => 1,
                    documentation => q{the max lane number on a flowcell},);

sub _build_max_lane {
  my ($self) = @_;
  return $self->general_values_conf()->{max_illumina_lanes};
}

=head2 owning_group - the owning group and files and folders should be made to be, defaults to value for group in general_values.ini

  my $sOwningGroup = $class->owning_group();

=cut

has q{owning_group} => (isa => q{Str}, is => q{ro}, lazy_build => 1,
                        documentation => q{the owning group for any files and folders created},);

sub _build_owning_group {
  my ($self) = @_;
  return $self->general_values_conf()->{group};
}

=head2 mail_from

the mail from to be used, can be set on construction, or defaults to value for default_mail_from in general_values.ini

  my $sMailFrom = $class->mail_from();

=cut

has q{mail_from} => (isa => q{Str}, is => q{ro}, lazy_build => 1,
                     documentation => q{the address of the email sender},);

sub _build_mail_from {
  my ($self) = @_;
  return $self->general_values_conf()->{default_mail_from};
}

=head2 control_lane_position - the position of the control lane on the flowcell, can be set on construction
or returns either the control lane from the npg::api::run object if you have the method run, returning this
or defaults to the value for preferred_control_lane from general_values.ini

  my $iControlLanePosition = $class->control_lane_position();

=cut

has q{control_lane_position} => (isa => q{Int}, is => q{ro}, lazy_build => 1,
                                 documentation => q{the position of the control lane},);

sub _build_control_lane_position {
  my ($self) = @_;
  my $control_lane = $self->_get_control_lane();
  if ($control_lane) {
    return $control_lane;
  }
  return $self->_default_control_lane();
}

############
# some processes need the control lane
#
sub _get_control_lane {
  my ($self) = @_;

  my $run;
  my $run_lanes;
  eval {
    $run = $self->run();
    $run_lanes = $run->run_lanes();
  } or do {}; # don't make fatal - this is only an option if a run object with run lanes can be found

  if (!$run_lanes) {
    return;
  }

  my $control_lane;

  foreach my $cont ( @{$run_lanes} ) {
    if ($cont->control()) {
      my ($pos) = $cont->position() =~ /(\d+)/xms;
      if ( ! $control_lane || ( $pos == $self->control_lane_position() ) ) {
        $control_lane = $pos;
      }
    }
  }

  return $control_lane;
}



=head2 default_control_lane

method returns the preferred_control_lane value from general_values.ini
for almost all situations, you are better off using control_lane_position

  my $iDefaultControlLane = $class->default_control_lane();

=cut

sub default_control_lane {
  my ($self) = @_;
  return $self->general_values_conf()->{preferred_control_lane};
}

=head2 internal_script_bin

on construction, set the script bin which should be used for scripts,
defaults to bin value set filesystem_locations.ini

=head2 default_internal_script_bin

method which could be used to set a script bin with a default set to bin value
set filesystem_locations.ini, if you have set internal_script_bin to something else

=cut

has q{internal_script_bin} => (isa => q{Str}, is => q{ro}, lazy_build => 1,
                               documentation => q{set the internal script bin where you expect to find the scripts},);

sub _build_internal_script_bin {
  my ($self) = @_;
  return $self->default_internal_script_bin();
}

sub default_internal_script_bin {
  my ($self) = @_;
  return $self->filesystem_locations_conf()->{bin};
}

=head2 bsub_error_file - returns a string to be used to for the error file, with 'substitute_this' in it to be swapped out for something unique,
can be set on construction, but you should include 'substitute_this' so that your file names can be made unique by the code

  my $sBsubErrorFile = $class->bsub_error_file();

=head2 bsub_output_file - returns a string to be used to for the output file, with 'substitute_this' in it to be swapped out for something unique,
can be set on construction, but you should include 'substitute_this' so that your file names can be made unique by the code

  my $sBsubOutputFile = $class->bsub_output_file();

=cut

has q{bsub_error_file}  => (isa => q{Str}, is => q{ro}, lazy_build => 1);
has q{bsub_output_file} => (isa => q{Str}, is => q{ro}, lazy_build => 1);

sub _build_bsub_error_file {
  my ($self) = @_;
  return $BSUB_ERROR_FILE;
}

sub _build_bsub_output_file {
  my ($self) = @_;
  return $BSUB_OUTPUT_FILE;
}

=head2 md5_creator_script

returns the script needed to be used for creating md5sums of files

=cut

sub md5_creator_script {
  return $MD5_CREATOR_SCRIPT;
}

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose::Role

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
