#############
# $Id: tag_info.pm 12818 2011-03-16 12:53:22Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-12-02
# Last Changed On: $Date: 2011-03-16 12:53:22 +0000 (Wed, 16 Mar 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-26.1/lib/npg_pipeline/roles/business/tag_info.pm $

package npg_pipeline::roles::business::tag_info;
use Moose::Role;
use Carp;
use English qw{-no_match_vars};
use Perl6::Slurp;

use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12818 $ =~ /(\d+)/mxs; $r; };

# most of the time, one of these defaults will be selected, based on the number of Index bases found
# leaving this in instead of moving to config file for now, since in general, most tags should now
# be being provided by sequencescape info
Readonly::Scalar our $TAGFILES      => {
    6 => q{/software/solexa/etc/barcodes/illumina12.tags},
    8 => q{/software/solexa/etc/barcodes/sanger.tags},
    7 => q{t/barcodes/sanger.tags},
  };


=head1 NAME

npg_pipeline::roles::business::tag_info

=head1 VERSION

$LastChangedRevision: 12818 $

=head1 SYNOPSIS

  package MyPackage;
  use Moose;
  ...
  with qw{npg_pipeline::roles::business::tag_info};

=head1 DESCRIPTION

This role exposes some methods to your class to deal with tag_files for multiplexing samples which describe the index 'barcode'.

=head1 SUBROUTINES/METHODS

=head2 tag_file

Attribute which contains the tag_file to be used. This is read/write, but should you require to set it manually, you would be
best to do so on construction, to ensure it is always correct when you call it.

  my $sTagFile = $class->tag_file();
  $class->tag_file( q{/Full/Path/And/Name/For/Tag/File} );

Ideally, for obtaining the file, you are best off using get_tag_file

=head2 has_tag_file

Returns a boolean based on if the tag_file has been set. Again, you are still better off using get_tag_file overall, unless
you cannot work out a number of index bases.

  my $bHasTagFile = $class->has_tag_file();

=cut

has q{tag_file} => (isa => q{Str}, is => q{rw}, predicate => q{has_tag_file}, clearer => q{_clear_tag_file},
  documentation => q{full path and name for tag file to be used for multiplexed samples, if you do not want a default, or lane specific file},);


=head2 get_tag_file

Separate reader method, which can take an argument of the index length, so if tag_file is not set, one of the defaults
can be used, or try to use a lane specific calculated index file.

  eval {
    my $sGetTagFile = $class->get_tag_file({
      index_length => $iIndexLength,
      tag_file_dir => $sTagFileDir, # Defaults to runfolder_path method
    });
  } or do {
    Your Error Handling Here...
  };

This method will croak if you do not provide 

i) index length, and no tag_file is provided or found, or the index length does not match one in the defaults
ii) tag_file_dir, and no runfolder_path method located (unless tag_file has been provided)
iii) lane ( $class->lane() ) or no lane specific tagfile can be found

It's preferential choice is

i)  declared tag_file given on object construction
ii) lane specific file which should be tag_file_dir/lane_<$lane>.tag (default runfolder_path/lane_<$lane>.tag)
iii) default chosen by index length

=cut

sub get_tag_file {
  my ($self, $arg_refs) = @_;

  # allow for possibility that just the index_length as an integer will be passed in, or that no hashref arg will be passed in
  $arg_refs = ref$arg_refs eq q{HASH}                 ? $arg_refs
            : $arg_refs && $arg_refs =~ /\A\d+\z/xms  ? { index_length => $arg_refs }
            :                                           {}
            ;

  my $index_length = $arg_refs->{index_length};

  # use a specified tag file to override any other file
  if ($self->has_tag_file()) {
    return $self->tag_file();
  }

  # use lane specific tag file constructed from location plus /lane_<$lane>.tag
  my $tag_file_dir = $arg_refs->{tag_file_dir} || $self->runfolder_path();
  my $lane = $self->lane() || q{};
  my $tag_file_name = $tag_file_dir . q{/lane_} . $lane . q{.tag};

  if ( -e $tag_file_name ) {
    return $tag_file_name;
  }

  # croak if no index length has been provided, as we cannot get a default

  if (!$index_length) {
    croak q{You have not provided an index length and there is no preset tag_file or lane specific tag file found.};
  }

  # use a default for this size of tag
  my $tag_file = $self->default_tag_files()->{$index_length};

  if (!$tag_file) {
    croak qq{There is no default tag_file for index length $index_length};
  }

  # check that the default tag_file exists and has tags in it, else croak out
  eval {
    $tag_file = $self->_check_legitimate_default($tag_file);
  } or do {
    croak $EVAL_ERROR;
  };

  return $tag_file;
}

=head2 default_tag_files

Returns a hashref of the default files, the key is the index length

  my $hDefaultTagFiles = $class->default_tag_files();

=cut

sub default_tag_files {
  return $TAGFILES;
}

=head2 corrected_tag_part_use_bases

this returns a corrected use_bases string, which will match the tag_file used for the lane.
Utilises get_tag_file to obtain the filename.
Your class must provide a use_bases method, for this to replace the index/multiplex part,
and an index_length method.

  eval {
    my $sCorrectedTagPartUseBases = $class->corrected_tag_part_use_bases({
      tag_file_dir => $sTagFileDir, # Defaults to runfolder_path method
    });
  } or do {
    Your Error Handling Here...
  };

=cut

sub corrected_tag_part_use_bases {
  my ($self, $arg_refs) = @_;
  $arg_refs ||= {};

  my $use_bases;
  eval {
    $use_bases = $self->use_bases();
  } or do {
    croak qq{This class object has no use_bases method:\n\n$EVAL_ERROR};
  };

  # non-multiplexed runs should just use $use_bases
  if (!$self->is_indexed()) {
    $self->_carp_it( qq{not indexed : $use_bases} );
    return $use_bases;
  }

  my $index_length;
  my $tag_file;

  eval {

    $index_length = $self->index_length();

  } or do {
    croak qq{This class object has no index_length method:\n\n$EVAL_ERROR};
  };

  my @parts = split /,/xms, $use_bases;

  my $index_read_n_count = $self->n_count($parts[1]);
  my $total_index_read_length = $index_length + $index_read_n_count;

  $arg_refs->{index_length} = $index_length;

  # try to retrieve a file tag_length. If the first default file can't be found from standard use_bases,
  # then try one which matches the full length of the index read
  my $tag_length;
  eval {

    $tag_length = $self->_retrieve_file_tag_length($arg_refs);
    1;
  } or do {

    $self->_carp_it( qq{Couldn't find a tag file for index_length $index_length, trying with total_index_read_length, $total_index_read_length} );

    $arg_refs->{index_length} = $total_index_read_length;

    eval {
      $tag_length = $self->_retrieve_file_tag_length($arg_refs);
      1;
    } or do {
      croak q{unable to obtain a file for either tag_length};
    };

  };

  if ($tag_length == $index_length) {
    return $use_bases;
  }

  if ($tag_length > $total_index_read_length) {
    croak qq{tag_lengths from file ($tag_length) > than index_read_length ($total_index_read_length)};
  }

  my $number_of_ns_needed = $total_index_read_length - $tag_length;

  my $n_part = $number_of_ns_needed == 1 ?  q{n}
             : $number_of_ns_needed > 1  ? qq{n$number_of_ns_needed}
             :                              q{}
             ;

  $parts[1] = q{I} . $tag_length . $n_part;

  return join q{,}, @parts;
}

=head2 n_count

counts and returns the number of n bases in the part of the use_bases string looked at

  my $iNCount = $class->n_count(q{I6nX});
  my $iNCount = $class->n_count(q{Y54nX});

=cut

sub n_count {
  my ($self, $part) = @_;

  my ($n_count) = $part =~ /n(\d+)/ixms;

  # if not then count the number of n's found
  if (!$n_count) {
    $n_count = $part =~ tr/Nn/Nn/;
  }

  return $n_count;
}

###############
# private methods

# does a count on the number of ACGT's found in the tag files first tag
sub _retrieve_file_tag_length {
  my ($self, $arg_refs) = @_;

  my $tag_file = $arg_refs->{tag_file} || $self->get_tag_file($arg_refs);
  my @tags = slurp $tag_file;
  my $tag_length = $tags[0] =~ tr/ACGT/ACGT/;

  return $tag_length;
}

# check that a given tag_file exists, can be read, and has at least 1 tag in it
sub _check_legitimate_default {
  my ($self, $tag_file) = @_;
  eval {
    $self->_retrieve_file_tag_length( { tag_file => $tag_file } );
  } or do {
    croak q{no legitimate default file found};
  };
  return $tag_file;
}

# _carp_it uses $self->log if it can, else carps
sub _carp_it {
  my ($self, $msg) = @_;
  if ($self->can(q{log})) {
    $self->log($msg);
    return;
  }

  carp $msg;
  return;
}

1;
__END__

=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose::Role

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
