#############
# $Id: base.pm 12818 2011-03-16 12:53:22Z ajb $
# Created By: ajb
# Last Maintained By: $Author: ajb $
# Created On: 2009-10-28
# Last Changed On: $Date: 2011-03-16 12:53:22 +0000 (Wed, 16 Mar 2011) $
# $HeadURL: svn+ssh://svn.internal.sanger.ac.uk/repos/svn/new-pipeline-dev/npg-pipeline/tags/release-26.1/lib/npg_pipeline/roles/generic/base.pm $

package npg_pipeline::roles::generic::base;
use Moose::Role;
use POSIX qw(strftime);
use English qw{-no_match_vars};
use Readonly; Readonly::Scalar our $VERSION => do { my ($r) = q$LastChangedRevision: 12818 $ =~ /(\d+)/mxs; $r; };
use Carp;
use File::Slurp;
use npg_pipeline::ConfigReader;

# change this to the location of your config directory if you need a different default location
Readonly::Scalar our $DEFAULT_CONF_DIR              => q{/software/solexa/src/new-pipeline-dev/npg-pipeline/trunk/data/config_files};

=head1 NAME

npg_pipeline::roles::generic::base

=head1 VERSION

$LastChangedRevision: 12818 $

=head1 SYNOPSIS

  package MyPackage;
  use Moose;
  ...
  with qw{npg_pipeline::roles::generic::base};

=head1 DESCRIPTION

This role provides some generic attributes and methods which are likely to be wanted for running a pipeline

With a few minor exceptions, everything inherits from npg_pipeline::base, so they will all have access to the
methods in this role.

This is why config reading is done in here.

=head1 SUBROUTINES/METHODS

=head2 script_name

returns the current scripts name (same as $PROGRAM_NAME), unless you have a desire to set on construction

  my $sScriptName = $class->script_name();

=cut

has q{script_name} => (isa => q{Str}, is => q{ro}, lazy_build => 1, documentation => q{Advise - Do Not set on construction - unless you need to, it can be built from $PROGRAM_NAME}); ## no critic (ValuesAndExpressions::RequireInterpolationOfMetachars)

sub _build_script_name {
  my ($self) = @_;
  return $PROGRAM_NAME;
}

=head2 time

returns and store localtime so that this can be used for calculations

  my $iTime = $class->time();

=head2 timestamp

returns and stores a timestring YYYY-MM-DD HH:MM:SS of $self->time()

  my $sTimeStamp = $class->timestamp();

=cut

has q{time}               => (isa => q{Int}, is => q{ro}, lazy_build => 1, documentation => q{Advise - Do Not set on construction, it can be built});
has q{timestamp}          => (isa => q{Str}, is => q{ro}, lazy_build => 1, documentation => q{Advise - Do Not set on construction, it can be built});

sub _build_time {
  my ($self) = @_;
  return time;
}

sub _build_timestamp {
  my ($self) = @_;
  my $ts = strftime '%Y%m%d-%H%M%S', localtime $self->time();
  return $ts;
}

=head2 lsf_queue

can be provided by the user, or will take default from either LSB_DEFAULTQUEUE environment variable or value for default_lsf_queue in general_values.ini

=cut

has q{lsf_queue}      => (isa => q{Str},  is => q{ro}, lazy_build => 1,
                          documentation => q{the lsf_queue you want to submit ordinary jobs to. defaults to LSB_DEFAULTQUEUE or from config file},);

sub _build_lsf_queue {
  my ($self) = @_;
  my $queue = $ENV{LSB_DEFAULTQUEUE} || $self->general_values_conf()->{default_lsf_queue};
  return $queue;
}

=head2 small_lsf_queue

can be provided by the user, or will take default from either LSB_DEFAULTQUEUE environment variable or value for default_lsf_queue in general_values.ini

=cut

has q{small_lsf_queue}  => (isa => q{Str},  is => q{ro}, lazy_build => 1,
                          documentation => q{the lsf_queue you want to submit small jobs to. defaults to value from config file},);

sub _build_small_lsf_queue {
  my ($self) = @_;
  return $self->general_values_conf()->{small_lsf_queue};
}

=head2 verbose

boolean to switch on verbose mode

=cut

has q{verbose}        => (isa => q{Bool},  is => q{ro}, documentation => q{Boolean decision to switch on verbose mode});

=head2 script_bin

attribute to store location of scripts to be run, if not set, then will default to an empty string

  my $sScriptBin = $class->script_bin();

=head2 set_script_bin

writer method to set the script_bin

  $class->set_script_bin($sScriptBin);

=cut

has q{script_bin} => (isa => q{Str}, is => q{ro}, writer => q{set_script_bin}, lazy_build => 1,
                      documentation => qq{attribute to store location of scripts to be run, if not set, then will default to an empty string\nafter construction, can be set with set_script_bin},);

sub _build_script_bin {
  my ($self) = @_;
  return q{};
}

=head2 use_bases

returns a string suitable for various scripts which shows the number of bases in order of sequencing eg Y54,I6n,y54
an implementation of _build_use_bases needs to be provided for this to work, no provision of this will cause a run_time error,
unless the string is provided on construction

  my $sUseBases = $class->use_bases();

=cut

# a writer has been provided for test purposes
has q{use_bases} => (isa => q{Str}, is => q{ro}, lazy_build => 1,  writer => q{_test_use_bases},
                     documentation => q{string which determines the bases in each reads (i.e. Y54,I6n,y54)},);


=head2 lane

option to push through for a single lane of a run only

=cut

has q{lane} => (isa => q{Int}, is => q{ro}, predicate => q{has_lane}, writer => q{_set_lane}, documentation => q{option to push through for a single lane of a run only});

before q{lane} => sub {
  my ($self) = @_;
  if (!$self->has_lane() && !$self->no_lanes()) {
    if ($self->count_lanes() == 1) {
      $self->_set_lane($self->first_lane());
    } else {
      carp q{no single lane given in lanes, so no single lane used};
    }
  }
};

=head2 lanes

option to push through an arrayref of lanes to run on

=head2 all_lanes

returns an array of the elements in $class->lanes();

=head2 no_lanes

returns true if no lanes have been specified

=head2 count_lanes

returns the number of lanes in $class->lanes()

=head2 first_lane

'shift's $class->lanes(), returning the element shifted
=cut

has q{lanes} => (
  traits        => ['Array'],
  isa           => q{ArrayRef[Int]},
  is            => q{ro},
  predicate     => q{has_lanes},
  documentation => q{option to push through selected lanes of a run only},
  default       => sub { [] },
  handles       => {
    all_lanes   => q{elements},
    no_lanes    => q{is_empty},
    count_lanes => q{count},
    first_lane  => q{shift},
  },
);

=head2 directory_exists

returns a boolean true or false dependent on the existence of directory

  my $bDirectoryExists = $class->directory_exists($sDirectoryPath);

=cut

sub directory_exists {
  my ($self, $directory_path) = @_;
  if (-d $directory_path) {
    return 1;
  }
  return 0;
}

=head2 lsb_jobindex

returns a useable string which can be dropped into the command which will be launched in the bsub job, where you
need $LSB_JOBINDEX, as this doesn't straight convert if it is required as part of a longer string

=cut

sub lsb_jobindex {
  return q{`echo $}. q{LSB_JOBINDEX`};
}

=head2 fs_resource_string

returns a resource string for the bsub command in format

  -R 'select rusage[nfs_sf=8]'
  -R 'select[nfs_12>=0] rusage[nfs_sf=8]' # we would like to include this, but it doesn't work with lsf6.1

optionally, can take a hashref which contains a resource string to modify and a value to use for the resource counter and
number of slots it will take (for example the number of processors)

  my $sSfResourceString = $oClass->fs_resource_string( {
    total_counter => 56, # defaults to 72 - doesn't work with lsf6.1, so don't bother
    counter_slots_per_job => 4, # defaults to 8
    resource_string => q{-R 'select[mem>8000] rusage[mem=8000] span[hosts=1]'}
  } );

=cut

sub fs_resource_string {
  my ( $self, $arg_refs ) = @_;
  my $resource_string = $arg_refs->{resource_string} || q{-R 'rusage[]'}; # q{-R 'select[] rusage[]'}; for when we can get a differen version of lsf

#  my ( $select ) = $resource_string =~ /select\[(.*?)\]/xms;
#  $select ||= q{};
#  my $new_select = $select;
  my ( $rusage ) = $resource_string =~ /rusage\[(.*?)\]/xms;
  $rusage ||= q{};
  my $new_rusage = $rusage;

#  if ( $new_select ) {
#    $new_select .= q{,};
#  }
#  $new_select .= $self->fs_resource() . q{>=} . ( $self->general_values()->{max_resource_counter_on_nodes} - ( $arg_refs->{total_counter} || $self->general_values()->{default_max_resource_counter} ) );

  if ( $new_rusage ) {
    $new_rusage .= q{,};
  }
  $new_rusage .= $self->fs_resource() . q{=} . ( $arg_refs->{counter_slots_per_job} || $self->general_values()->{default_resource_slots} );

#  $resource_string =~ s/select\[${select}\]/select[${new_select}]/xms;
  $resource_string =~ s/rusage\[${rusage}\]/rusage[${new_rusage}]/xms;

  return $resource_string;
}

###############
# config files

=head2 external_script_names_conf
=head2 filesystem_locations_conf
=head2 function_orders_conf
=head2 general_values_conf
=head2 illumina_pipeline_conf
=head2 pb_cal_pipeline_conf
=head2 references_conf
=head2 source_files_conf
=head2 emails_conf

All of these are accessors which return a hashref of configuration details for the current or declared domain
from the relevant configuration file

=cut

has [ qw{
    external_script_names_conf
    filesystem_locations_conf
    function_order_conf
    general_values_conf
    illumina_pipeline_conf
    pb_cal_pipeline_conf
    references_conf
    source_files_conf
    emails_conf
    parallelisation_conf
  } ] => (
  isa => q{HashRef},
  is  => q{ro},
  lazy_build => 1,
  init_arg => undef,
);

sub _build_external_script_names_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{external_script_names.ini} );
}

sub _build_filesystem_locations_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{filesystem_locations.ini} );
}

sub _build_function_order_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{function_orders.yml} );
}

sub _build_general_values_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{general_values.ini} );
}

sub _build_illumina_pipeline_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{illumina_pipeline.ini} );
}

sub _build_pb_cal_pipeline_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{pb_cal_pipeline.ini} );
}

sub _build_references_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{references.ini} );
}

sub _build_source_files_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{source_files.ini} );
}

sub _build_emails_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{emails.ini} );
}

sub _build_parallelisation_conf {
  my ( $self ) = @_;
  return $self->_get_config_reader( q{parallelisation.yml} );
}

sub _get_config_reader {
  my ( $self, $conf_name ) = @_;
  return npg_pipeline::ConfigReader->new( {
    conf_name => $conf_name,
    conf_path => $self->conf_path(),
    domain    => $self->domain(),
  } )->file_config();
}

=head2 domain

accessor to obtain the domain which should be looked for in the config files

=cut

has q{domain} => (
  isa => q{Str},
  is  => q{ro},
  lazy_build => 1,
  documentation => q{domain to utilise for config files. Defaults to $} . q{ENV{TEST} or live},
);

sub _build_domain {
  my ( $self ) = @_;
  return $ENV{TEST} || q{live};
}

has q{conf_path} => (
  isa => q{Str},
  is  => q{ro},
  default => $DEFAULT_CONF_DIR,
  documentation => qq{full path to directory of config files. Default: $DEFAULT_CONF_DIR},
);

=head2 run_conf

return a hash of the key value pairs in the current timestamps conf from the runfolder
useful for passing stuff between functions if there is a possibility that some info could do with
just being passed from one to another (to save extra odd lookups which could produce different data)

  my $hRunConf = $oClass->run_conf()

=cut

has q{run_conf} => (
  isa => q{HashRef},
  is  => q{ro},
  lazy_build => 1,
  init_arg => undef,
);

sub _build_run_conf {
  my ( $self ) = @_;

  my $run_folder    = $self->runfolder_path();               # fully qualified name

  my $file_config = {};
  my $conf_file = $self->timestamp() . q{.ini};
  if ( -e $run_folder . q{/} . $conf_file ) {
    $file_config = npg_pipeline::ConfigReader->new({
      conf_path => $run_folder,
      conf_name => $conf_file,
      domain    => $self->domain(),
    })->file_config();
  }
  return $file_config;
}

=head2 add_to_run_conf

add a/many key value pair(s) to the run conf file

  $oClass->add_to_run_conf( {
    key => value,
  } );

To add more, just add more key value pairs

returns a true value

=cut

sub add_to_run_conf {
  my ( $self, $args ) = @_;

  if ( ! ref $args || ref $args ne q{HASH} ) {
    croak qq{$args is not a HashRef};
  }

  my $file_config = $self->run_conf();

  my @lines = ( q{[} . $self->domain . qq{]\n} );

  # this has the added bonus of not needing to clear run_conf
  foreach my $key ( keys %{ $args } ) {
    $file_config->{$key} = $args->{$key};
  }

  foreach my $key ( sort keys %{ $file_config } ) {
    push @lines, qq{$key=$file_config->{$key}\n};
  }

  write_file( $self->runfolder_path() . q{/} . $self->timestamp() . q{.conf}, @lines );
  return 1;
}

=head2 fix_broken_files

Boolean flag to tell the pipeline to fix any missing broken files it can

=cut

has q{fix_broken_files} => (
  isa => q{Bool},
  is  => q{ro},
  documentation => q{boolean flag to fix files that may be missing or broken if it can},
);

=head2 make_log_dir

creates a log_directory in the given directory

  $oMyPackage->make_log_dir( q{/dir/for/base} );

=cut

sub make_log_dir {
  my ( $self, $dir, $owning_group ) = @_;

  my $log_dir = qq{$dir/log};

  $owning_group ||= $ENV{OWNING_GROUP};
  $owning_group ||= $self->general_values_conf()->{group};

  if ( -d $log_dir ) {
    if ( $self->can( q{log} ) ) {
      $self->log( qq{$log_dir already exists} );
    }
    return $log_dir;
  }

  my $cmd = qq{mkdir -p $log_dir};
  my $output = qx{$cmd};

  if ( $self->can( q{log} ) ) {
    $self->log( qq{Command: $cmd} );
    $self->log( qq{Output:  $output} );
  }

  if ( $CHILD_ERROR ) {
    croak qq{unable to create $log_dir:$output};
  }

  if ( $self->can( q{log} ) ) {
    $self->log( qq{chgrp $owning_group $log_dir} );
  }

  my $rc = qx{chgrp $owning_group $log_dir};
  if ( $CHILD_ERROR ) {
    if ( $self->can( q{log} ) ) {
      $self->log("could not chgrp $log_dir\n\t$rc");                # not fatal
    }
  }
  $rc = qx{chmod u=rwx,g=srxw,o=rx $log_dir};
  if ( $CHILD_ERROR ) {
    $self->log("could not chmod $log_dir\n\t$rc");                # not fatal
  }

  return $log_dir;
}

=head2 job_name_prefix

value to be prepended to job_names to signify something about where they have been launched from (i.e. prod for production). No need to use an underscore in the prefix, this will be added automatically

This can be set in the general_values.ini config file, but will be overridden if given on the command line.

=cut

has q{job_name_prefix_store} => (
  isa => q{Str},
  is  => q{ro},
  init_arg => q{job_name_prefix},
  documentation => q{give all your jobs a prefix to their names},
  predicate => q{has_job_name_prefix_store},
);

sub job_name_prefix {
  my ( $self ) = @_;

  my $job_name_prefix = $self->has_job_name_prefix_store()              ? $self->job_name_prefix_store() . q{_}
                      : $self->general_values_conf()->{job_name_prefix} ? $self->general_values_conf()->{job_name_prefix} . q{_}
                      :                                                   q{}
                      ;

  return $job_name_prefix;
}

1;
__END__


=head1 DIAGNOSTICS

=head1 CONFIGURATION AND ENVIRONMENT

=head1 DEPENDENCIES

=over

=item Moose::Role

=item Carp

=item English -no_match_vars

=item Readonly

=back

=head1 INCOMPATIBILITIES

=head1 BUGS AND LIMITATIONS

=head1 AUTHOR

$Author: ajb $

=head1 LICENSE AND COPYRIGHT

Copyright (C) 2010 GRL, by Andy Brown (ajb@sanger.ac.uk)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
